package {
	
	import flash.display.Sprite;
	import flash.display.DisplayObject;
	import flash.display.Bitmap;
	import flash.text.TextField;
	import flash.display.StageScaleMode;
	import flash.display.StageQuality;
	import flash.utils.Dictionary;
	import flash.events.Event;
	import flash.events.TimerEvent;
	import flash.events.KeyboardEvent;
	import flash.geom.Point;
	import flash.utils.Timer;

	[SWF(width="160", height="160", backgroundColor="#302b07")] //#004012

	public class Main extends Sprite {

		[Embed(source = '../assets/spritesheet_yellow.png')]
		public static var SpriteSheet: Class;
		public static var debug: Debug;
		public static var spriteSheet: Bitmap;

//		public static var FPS: int = 10;
		public static var TILE_SIZE: int = 16;
		public static var GAME_WIN: int = 1;
		public static var GAME_LOSE: int = 0;
		public static var GAME_PLAYING: int = 3;
		public static var GAME_WAITING: int = -1;
		public static var MOVE_SPEED: Number = TILE_SIZE/8;

		private var _map1: Map;
		private var _map2: Map;
		private var _train: Train;
		private var _playerChar: PlayerCharacter;
		private var _title: Title;
		private var _win: Win;
		private var _lose: Lose;
		private var _clock: Clock;

		private var _10SecTimer: Timer;
		private var _startDelay: Timer;

		private var _keyUp: Boolean = false;
		private var _keyDown: Boolean = false;
		private var _keyLeft: Boolean = false;
		private var _keyRight: Boolean = false;

		private var _state: int = GAME_WAITING;
		private var _startPos: Point;

		private var _people: Array;

		public function Main() {
			stage.scaleMode = StageScaleMode.SHOW_ALL;
			stage.quality = StageQuality.LOW;
//			stage.frameRate = FPS;
	
			debug = new Debug(0x0000ff);
			spriteSheet = new SpriteSheet() as Bitmap;

			this._startPos = new Point(-TILE_SIZE, -9*TILE_SIZE);

			// Init timers
			this._startDelay = new Timer(2000, 1);
			this._startDelay.addEventListener(TimerEvent.TIMER_COMPLETE, this._startGame);
			this._10SecTimer = new Timer(1000, 10);
			this._10SecTimer.addEventListener(TimerEvent.TIMER, this._onTick);
			this._10SecTimer.addEventListener(TimerEvent.TIMER_COMPLETE, this._onTimerComplete);
			
			// Init game objects
			_initGameObjects();

			// Add events
			this.addEventListener(Event.ENTER_FRAME, _update);
			stage.addEventListener(KeyboardEvent.KEY_DOWN, _onKeyDown);
			stage.addEventListener(KeyboardEvent.KEY_UP, _onKeyUp);

			addChild(debug); // DEBUG

			this._startDelay.start();
		}

		private function _initGameObjects():void {
			// Map Layer One
			var mapData1: Array = [
				[15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15],
				[15, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 15],
				[15, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 1, 0, 15],
				[15, 0, 0, 0, 1, 0, 1, 0, 1, 0, 0, 0, 0, 15],
				[15, 9, 9, 8, 8, 8, 8, 8, 8, 8, 8, 9, 9, 15],
				[15, 10, 10, 15, 15, 15, 15, 15, 15, 15, 15, 10, 10, 15]
			];
			_map1 = new Map(mapData1, this._startPos.x, -(5*TILE_SIZE)+this._startPos.y);
			addChild(_map1);

			// Train
			_train = new Train(0, 0, 2);
			addChild(_train);

			// Map Layer Two
			var mapData2: Array = [
				[15, 15, 15, 5, 5, 5, 5, 5, 5, 5, 5, 15, 15, 15],
				[15, 0, 1, 6, 6, 6, 6, 6, 6, 6, 6, 0, 0, 15],
				[15, 0, 0, 2, 2, 2, 2, 2, 2, 2, 2, 0, 0, 15],
				[15, 0, 0, 0, 1, 2, 2, 2, 2, 0, 1, 0, 0, 15],
				[15, 0, 0, 0, 0, 2, 2, 2, 2, 0, 0, 1, 0, 15],
				[15, 0, 0, 0, 0, 2, 2, 2, 2, 0, 0, 0, 0, 15],
				[15, 0, 1, 0, 0, 2, 2, 2, 2, 0, 1, 0, 0, 15],
				[15, 0, 0, 1, 0, 2, 2, 2, 2, 0, 0, 1, 0, 15],
				[15, 0, 0, 0, 0, 2, 2, 2, 2, 0, 0, 0, 0, 15],
				[15, 0, 0, 0, 0, 2, 2, 2, 2, 0, 1, 0, 0, 15],
				[15, 1, 0, 0, 0, 2, 2, 2, 2, 0, 0, 1, 0, 15],
				[15, 0, 0, 0, 0, 2, 2, 2, 2, 0, 0, 0, 0, 15],
				[15, 0, 0, 0, 0, 2, 2, 2, 2, 1, 0, 0, 0, 15],
				[15, 0, 0, 1, 0, 2, 2, 2, 2, 0, 0, 1, 0, 15],
				[15, 0, 0, 0, 0, 2, 2, 2, 2, 0, 0, 0, 0, 15],
				[15, 0, 0, 0, 0, 2, 2, 2, 2, 0, 0, 1, 0, 15],
				[15, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 15],
				[15, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 15],
				[15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15]
			];
			_map2 = new Map(mapData2,  this._startPos.x, this._startPos.y);
			addChild(_map2);

			// Add bouncy characters
			_people = this._addCharacters();

			// Player
			_playerChar = new PlayerCharacter(stage.stageWidth/2, stage.stageHeight/2);
			addChild(_playerChar);

			// Add the clock
			_clock = new Clock(0, 0);
			addChild(_clock);

			// Game state overlays/screens
			_win = new Win(stage.stageWidth/2, stage.stageHeight/2, false);
			addChild(_win);
			_lose = new Lose(stage.stageWidth/2, stage.stageHeight/2, false);
			addChild(_lose);
			_title = new Title(stage.stageWidth/2, stage.stageHeight/2);
			addChild(_title);
		}

		private function _addCharacters():Array {
			var chars:Array = [];
			var r:int = 9;
			var c:int = 5;
			for (var i:int=0;i<r;i++) {
				for (var f:int=0;f<c;f++) {
					var ch:CharWithHat = new CharWithHat(
						_map2, 
						f*TILE_SIZE+(c*TILE_SIZE-TILE_SIZE*1),
						i*TILE_SIZE-r*TILE_SIZE
					);
					addChild(ch);
					chars.push(ch);
				}
			}
			for (var k:int=0;k<r;k++) {
				for (var l:int=0;l<c;l++) {
					var h:CharWithHat = new CharWithHat(
						_map2, 
						l*TILE_SIZE+(c*TILE_SIZE-TILE_SIZE*4),
						k*TILE_SIZE-(r*TILE_SIZE+TILE_SIZE*6)
					);
					addChild(h);
					chars.push(h);
				}
			}
			for (var m:int=0;m<r;m++) {
				for (var n:int=0;n<c;n++) {
					var o:CharWithHat = new CharWithHat(
						_map2, 
						n*TILE_SIZE+(c*TILE_SIZE+TILE_SIZE*3),
						m*TILE_SIZE-(r*TILE_SIZE+TILE_SIZE*6)
					);
					addChild(o);
					chars.push(o);
				}
			}
			for (var p:int=0;p<r;p++) {
				for (var u:int=0;u<c;u++) {
					var q:CharWithHat = new CharWithHat(
						_map2, 
						u*TILE_SIZE+(c*TILE_SIZE-TILE_SIZE*1),
						p*TILE_SIZE-(r*TILE_SIZE+TILE_SIZE*8)
					);
					addChild(q);
					chars.push(q);
				}
			}
			return chars;
		}

		private function _startGame(e:TimerEvent):void {
			this._10SecTimer.start();
			this._state = GAME_PLAYING;
			this._title.visible = false;
		}

		private function _onTick(e:TimerEvent):void {
			this._clock.showFrame(_10SecTimer.currentCount-1<0?0:_10SecTimer.currentCount-1);
		}

		private function _onTimerComplete(e:TimerEvent):void {
			this._clock.showFrame(0);
			this.endGame(false);
		}

		private function _update(e:Event):void {
			_map1.update(e);
			_map2.update(e);
			_updateMoveOffset();
			if (_state == GAME_WIN) {
				_train.moveY(_map1.y+3*TILE_SIZE);
				_animateAlpha(_win, 1);
				_catchedTrainAnimation();
				_hideChars();
			} else if (_state == GAME_LOSE) {
				_train.moveY(_map1.y+3*TILE_SIZE);
				_animateAlpha(_lose, 1);
				_toLateAnimation();
				_hideChars();
			} else if (_state == GAME_PLAYING) {
				_updateChars(e);
				_playerChar.updateAnimation(_map1.vel);
				_checkCollision();
				if (_map1.x > Main.TILE_SIZE*4) {
					_map1.vel.x = _map2.vel.x = 0;
					_map1.x = _map2.x = Main.TILE_SIZE*4;
				}
				if (_map1.x < -Main.TILE_SIZE*7) {
					_map1.vel.x = _map2.vel.x = 0;
					_map1.x = _map2.x = -Main.TILE_SIZE*7;
				}
			}

			if (_state == GAME_PLAYING || _state == GAME_WAITING) {
				_train.moveTo(_map1.x+2*TILE_SIZE, _map1.y+3*TILE_SIZE);
			}
		}

		private function _updateMoveOffset():void {
			if (_keyUp) {
				_map1.velY(1);
				_map2.velY(1);
			} else if (_keyDown) {
				_map1.velY(-1);
				_map2.velY(-1);
			} else {
				_map1.velY(0);
				_map2.velY(0);
			}

			if (_keyLeft) {
				_map1.velX(1);
				_map2.velX(1);
			} else if (_keyRight) {
				_map1.velX(-1);
				_map2.velX(-1);
			} else {
				_map1.velX(0);
				_map2.velX(0);
			}
		}

		private function _updateChars(e:Event):void {
			for (var i:int=0;i<_people.length;i++) {
				_people[i].update(e);
				if (
					Point.distance(_people[i].getPos(), _playerChar.getPos())
					< TILE_SIZE/1.5
				) {
//					_people[i].vel.y = -_map2.vel.yj
//					_map2.vel.x = _map1.vel.x = -_people[i].vel.x;
					_map2.vel.y = _map1.vel.y = -_people[i].vel.y;
//					_people[i].vel.x = -_map2.vel.x;
//					_people[i].vel.y = -_map2.vel.y;
//					_people[i].vel = new Point(-pv.x, -pv.y);
//					_map2.vel = _map1.vel = new Point(-cv.x, -cv.y);
				}
			}
		}

		private function _checkCollision():void {
			// You won!
			if (this._playerChar.hitTestObject(this._train)) this.endGame(true);
		}

		private function _onKeyDown(e:KeyboardEvent):void {
			if (this._state == GAME_PLAYING) {
				if (e.keyCode == 87 || e.keyCode == 38) {
					_keyUp = true;
					_keyDown = false;
				} else if (e.keyCode ==  83 || e.keyCode == 40) {
					_keyDown = true;
					_keyUp = false;
				}
				if (e.keyCode == 65 || e.keyCode == 37) {
					_keyLeft = true;
					_keyRight = false;
				} else if (e.keyCode == 68 || e.keyCode == 39) {
					_keyRight = true;
					_keyLeft = false;
				}
			} else {
				_keyUp = _keyDown = _keyRight = _keyLeft = false;
			}
			if (e.keyCode == 82) {
				this._reset();
			}
		}

		private function _onKeyUp(e:KeyboardEvent):void {
			if (this._state == GAME_PLAYING) {
				if (e.keyCode == 87 || e.keyCode == 38) {
					_keyUp = false;
				} else if (e.keyCode ==  83 || e.keyCode == 40) {
					_keyDown = false;
				}
				if (e.keyCode == 65 || e.keyCode == 37) {
					_keyLeft = false;
				} else if (e.keyCode == 68 || e.keyCode == 39) {
					_keyRight = false;
				}
			} else {
				_keyUp = _keyDown = _keyRight = _keyLeft = false;
			}
		}

		public function endGame(wonGame:Boolean):void {
			if (_state == GAME_PLAYING) {
				this._state = (wonGame?GAME_WIN:GAME_LOSE);

				if (wonGame) _10SecTimer.stop();

				if (wonGame) this._playerChar.visible = false;
			}
		}

		private function _animateAlpha(o:DisplayObject, a:int, s:Number = .1):void {
			if (o.alpha != a) o.alpha += (a>0?s:-s);
			else o.alpha = a;
		}

		private function _catchedTrainAnimation():void {
			_map1.velX(1);
			_map2.velX(1);
			if (_map1.x > 10*16) {
				_animateAlpha(_train, 0, .05);
				_animateAlpha(_clock, 0, .05);
			}
		}

		private function _toLateAnimation():void {
			_train.x -= TILE_SIZE/8;
		}

		private function _reset():void {
			this._10SecTimer.stop();
			this._10SecTimer.reset();
			this._startDelay.reset();
			this._lose.alpha = this._win.alpha = 0;
			this._train.alpha = this._clock.alpha = 1;
			this._map2.x = this._map1.x = this._startPos.x;
			this._map2.y = this._startPos.y;
			this._map1.y = this._startPos.y-5*TILE_SIZE;
			this._resetChars();
			this._playerChar.visible = true;
			this._clock.showFrame(0);
			this._state = GAME_WAITING;
			this._startDelay.start();
		}

		private function _hideChars():void {
			for (var i:int=0;i<this._people.length;i++) {
				this._people[i].visible = false;
			}
		}

		private function _resetChars():void {
			for (var i:int=0;i<this._people.length;i++) {
				this._people[i].visible = true;
				this._people[i].resetPosition();
			}
		}

	}

}
